﻿#if DEBUG
    //#define DRAW_CLUSTERS_IN_CONSOLE
    //#define SHOW_TOT_VALS
    //#define DRAW_NEIGHBOURS
#endif

using System;
using System.Collections.Generic;
using System.Linq;

namespace Jacquot_Timepix_V2.Models
{
    public class Cluster
    {
        public uint ID { get; }
        public ushort X { get; }
        public ushort Y { get; }
        public uint Time { get; }
        public ushort Size { get; }
        public double Energy { get; }
        public ClusterType Physical { get; }
        public RadiationType Particle { get; }

        public List<Pixel> Pixels { get; }
        public double AverageTot { get; }
        public double FillRatio { get; }
        public double Linearity { get; }
        public double StandardDeviation { get; }

        /// <summary>
        /// Use <see cref="Create"/> method to generate Cluster objects
        /// </summary>
        private Cluster(
            uint id, ushort x, ushort y, uint time, ushort size, double energy,
            ClusterType physical,
            RadiationType particle,
            List<Pixel> pixels)
        {
            ID = id;
            X = x;
            Y = y;
            Time = time;
            Size = size;
            Energy = energy;

            Physical = physical;
            Particle = particle;

            Pixels = pixels;

            AverageTot = Pixels.Average(p => p.Tot);
            FillRatio = GetFillRatio();
            Linearity = GetLinearity();
            StandardDeviation = GetDeviation();
        }

        /// <returns>
        /// <see cref="Cluster"/> if everything is valid. Returns <see cref="null"/> if cluster isn't valid.
        /// </returns>
        public static Cluster? Create(
            uint id, ushort x, ushort y, uint time, ushort size, double energy,
            ClusterType physical,
            RadiationType particle,
            List<Pixel> pixels)
        {
            if (energy < 0)
                return null;

            // room for more properties to check

            return new(id, x, y, time, size, energy, physical, particle, pixels);
        }

        #region FillRatio
        /// <summary>
        /// <para>Updates <see cref="PixelType"/> <see cref="Pixel.Type"/> for every <see cref="Pixel"/></para>
        /// <para><see cref="Pixel.Type"/> properties are uninitialized before this method is called</para>
        /// </summary>
        /// <returns>
        /// inside pixels / total pixels
        /// </returns>
        private double GetFillRatio()
        {
            if (Size < Pixel.insidePixelNeighbours)
            {
                Pixels.ForEach(p => p.Type = PixelType.Edge);
                return 0;
            }

            // first find min/max of x/y
            int xMin = Pixels[0].X;
            int xMax = xMin;
            int yMin = Pixels[0].Y;
            int yMax = yMin;

            // create 2D bool array to store pixels
            for (int i = 1; i < Pixels.Count; i++)
            {
                int x = Pixels[i].X;
                int y = Pixels[i].Y;

                xMin = x < xMin ? x : xMin;
                xMax = x > xMax ? x : xMax;
                yMin = y < yMin ? y : yMin;
                yMax = y > yMax ? y : yMax;
            }
            int width = xMax - xMin + 1;
            int height = yMax - yMin + 1;

            // default byte value is 0x00
            byte[,] neighbours = new byte[width, height];

#if DRAW_CLUSTERS_IN_CONSOLE
            ushort totMax = 0;
            Pixel?[,] pixels = new Pixel?[width, height];
#endif

            // populate array
            foreach (Pixel pixel in Pixels)
            {
                int x = pixel.X - xMin;
                int y = pixel.Y - yMin;

                AddPixel(ref neighbours, x, y, width, height);

#if DRAW_CLUSTERS_IN_CONSOLE
                totMax = pixel.Tot > totMax ? pixel.Tot : totMax;
                pixels[x, y] = pixel;
#endif
            }

            // check neighbours
            int insidePixels = 0, edgePixels = 0;
            foreach (Pixel pixel in Pixels)
            {
                int x = pixel.X - xMin;
                int y = pixel.Y - yMin;

                if (neighbours[x, y] >= Pixel.insidePixelNeighbours)
                {
                    insidePixels++;
                    pixel.Type = PixelType.Inside;
                }
                else
                {
                    edgePixels++;
                    pixel.Type = PixelType.Edge;
                }
            }

#if DRAW_CLUSTERS_IN_CONSOLE
            // 5 levels of pixel brigthness
            char[] fill = new char[] { ' ', '░', '▒', '▓', '█', '█', '█' };
            double div = totMax / 5.0;
#if SHOW_TOT_VALS
            int chars = totMax.ToString().Length + 1;
#endif

            double fillRatio = ((double)insidePixels / (insidePixels + edgePixels));
            double linearity = GetLinearity();
            double variation = GetDeviation();

            // search for a specific cluster? use: fr, lin, vr, width, heigh, ID
            if (variation > 0.75 && variation < 1.5)
            {
                Console.WriteLine("\n--Cluster ID: " + ID + "\tX: " + X + "\tY: " + Y + "\tTime: " + Time + "\tSize: " + Size + "\tEnergy: " + Energy);
                for (int row = 0; row < height; row++)
                {
                    for (int col = 0; col < width; col++)
                    {
#if SHOW_TOT_VALS
                        Console.Write(pixels[col, row] is not null ? pixels[col, row]?.Tot.ToString().PadLeft(chars) : "0".PadLeft(chars));
#else
                        Console.Write(pixels[col, row] is not null ? fill[(int)(pixels[col, row].Tot / div + 1)] : fill[0]);
#endif
                    }
#if DRAW_NEIGHBOURS
                    Console.Write("  |  ");
                    for (int col = 0; col < width; col++)
                    {
                        Console.Write(neighbours[col, row] != 0 ? neighbours[col, row].ToString() : '-');
                    }
#endif
                    Console.Write('\n');
                }
                Console.WriteLine("Classified as physical: " + Physical + ", particle: " + Particle);
                Console.WriteLine("Fill ratio: " + fillRatio);
                Console.WriteLine("Linearity: " + linearity);
                Console.WriteLine("Variation: " + variation);
#if SHOW_TOT_VALS
                Console.WriteLine("Max Tot: " + totMax);
#endif
                _ = Console.ReadKey();
            }
#endif

                return ((double)insidePixels / (insidePixels + edgePixels));
        }

        /// <summary>
        /// Increments all of <c>grid[x,y]</c>'s neighbours by 1.
        /// </summary>
        private static void AddPixel(ref byte[,] neighbours, int x, int y, int width, int height)
        {
            // increment 4 surrounding pixels that share a side
            TryIncrement(ref neighbours, x, y - 1, width, height);
            TryIncrement(ref neighbours, x, y + 1, width, height);
            TryIncrement(ref neighbours, x - 1, y, width, height);
            TryIncrement(ref neighbours, x + 1, y, width, height);
            if (Pixel.checkCornerNeighbours)
            {
                // increment 4 surrounding pixels that share a corner
                TryIncrement(ref neighbours, x - 1, y - 1, width, height);
                TryIncrement(ref neighbours, x - 1, y + 1, width, height);
                TryIncrement(ref neighbours, x + 1, y - 1, width, height);
                TryIncrement(ref neighbours, x + 1, y + 1, width, height);
            }
        }

        /// <summary>
        /// Increments <c>grid[x,y]</c> by 1 if x and y are inside array bounds
        /// </summary>
        private static void TryIncrement(ref byte[,] array, int x, int y, int width, int height)
        {
            if (x >= 0 && x < width && y >= 0 && y < height)
            {
                array[x, y]++;
            }
        }
        #endregion

        #region Linearity
        /// <summary>
        /// Finds the maximum length between any 2 <see cref="Pixel"/>s in the entire <see cref="Cluster"/>
        /// </summary>
        /// <returns>max length / total pixels</returns>
        private double GetLinearity()
        {
            // TODO: optimize using "Rotating Calipers" Shamos algorithm

            List<Pixel> edgePixels = Pixels.FindAll(p => p.Type == PixelType.Edge);

            if (edgePixels.Count <= 1)
            {
                return 0;
            }

            List<Pixel> done = new();
            int lengthSquaredMax = 0;

            // checks every pair of 2 pixels exactly once
            foreach (Pixel pix1 in edgePixels)
            {
                done.Add(pix1);
                foreach (Pixel pix2 in edgePixels.Except(done))
                {
                    int dx = pix2.X - pix1.X;
                    int dy = pix2.Y - pix1.Y;

                    // instead of square rooting every loop, compare squared values and square root only once at the end
                    int lengthSquared = dx * dx + dy * dy;
                    lengthSquaredMax = lengthSquared > lengthSquaredMax ? lengthSquared : lengthSquaredMax;
                }
            }

            // square root
            double lengthMax = Math.Pow(lengthSquaredMax, 0.5);

            return lengthMax / Pixels.Count;
        }
        #endregion

        #region StandardDeviation
        private double GetDeviation()
        {
            if (Pixels.Count > 1)
            {
                double sum = Pixels.Select(p => (p.Tot - AverageTot) * (p.Tot - AverageTot)).Sum();
                return Math.Pow(sum / (Pixels.Count - 1), 0.5);
            }
            else
            {
                return 0;
            }
        }
        #endregion
    }

    /// <summary>
    /// Shape of the cluster
    /// </summary>
    public enum ClusterType
    {
        CURLY_TRACK,
        DOT,
        HEAVY_BLOB,
        HEAVY_TRACK,
        SMALL_BLOB,
        STRAIGHT_TRACK
    }

    public enum RadiationType
    {
        ALPHA,
        BETA,
        GAMA,
        UNDEFINED_PART
    }
}
